"""
Module provides classes for exporting SCADA data stored in
:class:`~epyt_flow.simulation.scada.scada_data.ScadaData`.
"""
from abc import abstractmethod
import numpy as np
from scipy.io import savemat
import pandas as pd

from .scada_data import ScadaData
from ..sensor_config import SensorConfig, massunit_to_str, flowunit_to_str, qualityunit_to_str, \
    is_flowunit_simetric


class ScadaDataExport():
    """
    Base class for exporting SCADA data stored in
    :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`.

    Parameters
    ----------
    f_out : `str`
        Path to the file to which the SCADA data will be exported.
    export_raw_data : `bool`, optional
        If True, the raw measurements (i.e. sensor reading without any noise or faults)
        are exported instead of the final sensor readings.

        The default is False.
    """

    def __init__(self, f_out: str, export_raw_data: bool = False, **kwds):
        self.__f_out = f_out
        self.__export_raw_data = export_raw_data

        super().__init__(**kwds)

    @property
    def f_out(self) -> str:
        """
        Gets the path to the file to which the SCADA data will be exported.

        Returns
        -------
        `str`
            Path to the file to which the SCADA data will be exported.
        """
        return self.__f_out

    @property
    def export_raw_data(self) -> bool:
        """
        True if the raw measurements instead of the final sensor readings are requested.

        Returns
        -------
        `bool`
            True if the raw measurements instead of the final sensor readings are requested.
        """
        return self.__export_raw_data

    @staticmethod
    def create_global_sensor_config(scada_data: ScadaData) -> SensorConfig:
        """
        Creates a global sensor configuration with sensors placed everywhere.

        Parameters
        ----------
        scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            SCADA data for which the global sensor configuration is to be created.

        Returns
        -------
        :class:`~epyt_flow.simulation.sensor_config.SensorConfig`
            Global sensor configuration.
        """
        old_sensor_config = scada_data.sensor_config

        sensor_config = SensorConfig.create_empty_sensor_config(old_sensor_config)
        sensor_config.pressure_sensors = sensor_config.nodes
        sensor_config.flow_sensors = sensor_config.links
        sensor_config.demand_sensors = sensor_config.nodes
        sensor_config.quality_node_sensors = sensor_config.nodes
        sensor_config.quality_link_sensors = sensor_config.links
        sensor_config.valve_state_sensors = sensor_config.valves
        sensor_config.tank_level_sensors = sensor_config.tanks
        sensor_config.pump_state_sensors = sensor_config.pumps
        sensor_config.bulk_species_node_sensors = sensor_config.bulk_species_node_sensors
        sensor_config.bulk_species_link_sensors = sensor_config.bulk_species_link_sensors
        sensor_config.surface_species_sensors = sensor_config.surface_species_sensors

        return sensor_config

    @staticmethod
    def create_column_desc(scada_data: ScadaData) -> np.ndarray:
        """
        Creates column descriptions -- i.e. sensor type and location for each column

        Parameters
        ----------
        scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            SCADA data to be described.

        Returns
        -------
        `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            3-dimensional array describing all columns of the sensor readings:
            The first dimension describes the sensor type, the second dimension
            describes the sensor location, and the third one describes the measurement units.
        """
        sensor_readings = scada_data.get_data()

        def __get_sensor_unit(sensor_type):
            if sensor_type == "pressure":
                if not is_flowunit_simetric(scada_data.sensor_config.flow_unit):
                    return "psi"
                else:
                    return "meter"
            elif sensor_type == "flow" or sensor_type == "demand":
                return flowunit_to_str(scada_data.sensor_config.flow_unit)
            elif sensor_type == "quality_node" or sensor_type == "quality_link":
                return qualityunit_to_str(scada_data.sensor_config.quality_unit)
            elif sensor_type == "tank_volume":
                if is_flowunit_simetric(scada_data.sensor_config.flow_unit):
                    return "cubic meter"
                else:
                    return "cubic foot"
            else:
                return ""

        col_desc = [None for _ in range(sensor_readings.shape[1])]
        sensor_config = scada_data.sensor_config
        sensors_id_to_idx = sensor_config.sensors_id_to_idx
        for sensor_type in sensors_id_to_idx:
            unit_desc = __get_sensor_unit(sensor_type)
            for item_id in sensors_id_to_idx[sensor_type]:
                col_id = sensors_id_to_idx[sensor_type][item_id]

                if sensor_type == "bulk_species_node" or sensor_type == "bulk_species_link":
                    bulk_species_idx = sensor_config.bulk_species.index(item_id)
                    unit_desc = massunit_to_str(sensor_config.
                                                bulk_species_mass_unit[bulk_species_idx])
                elif sensor_type == "surface_species":
                    surface_species_idx = sensor_config.surface_species.index(item_id)
                    unit_desc = massunit_to_str(sensor_config.
                                                bulk_species_mass_unit[surface_species_idx])

                if sensor_type not in ["bulk_species_node", "bulk_species_link", "surface_species"]:
                    col_desc[col_id] = [sensor_type, item_id, unit_desc]
                else:
                    for location_id, c_id in col_id.items():
                        col_desc[c_id] = [sensor_type, f"{item_id} @ {location_id}", unit_desc]

        return np.array(col_desc, dtype=object)

    @abstractmethod
    def export(self, scada_data: ScadaData) -> None:
        """
        Exports given SCADA data.

        Parameters
        ----------
        scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            SCADA data to be exported.
        """
        raise NotImplementedError()


class ScadaDataNumpyExport(ScadaDataExport):
    """
    Class for exporting SCADA data to numpy (.npz file).
    """

    def export(self, scada_data: ScadaData) -> None:
        """
        Exports given SCADA data.

        Parameters
        ----------
        scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            SCADA data to be exported.
        """
        if not isinstance(scada_data, ScadaData):
            raise TypeError("'scada_data' must be an instance of " +
                            "'epyt_flow.simulation.scada_data.ScadaData' and not of " +
                            f"'{type(scada_data)}'")

        old_sensor_config = None
        if self.export_raw_data is True:
            # Backup old sensor config and set a new one with sensors everywhere
            old_sensor_config = scada_data.sensor_config
            scada_data.change_sensor_config(self.create_global_sensor_config(scada_data))

        sensor_readings = scada_data.get_data()
        col_desc = self.create_column_desc(scada_data)
        sensor_readings_time = scada_data.sensor_readings_time

        if self.export_raw_data is True:
            # Restore old sensor config
            scada_data.change_sensor_config(old_sensor_config)

        np.savez(self.f_out, sensor_readings=sensor_readings, col_desc=col_desc,
                 sensor_readings_time=sensor_readings_time,
                 flow_unit=scada_data.sensor_config.flow_unit)


class ScadaDataXlsxExport(ScadaDataExport):
    """
    Class for exporting SCADA data to Excel (.xlsx file).
    """

    def export(self, scada_data: ScadaData) -> None:
        """
        Exports given SCADA data.

        Parameters
        ----------
        scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            SCADA data to be exported.
        """
        if not isinstance(scada_data, ScadaData):
            raise TypeError("'scada_data' must be an instance of " +
                            "'epyt_flow.simulation.scada_data.ScadaData' and not of " +
                            f"'{type(scada_data)}'")

        old_sensor_config = None
        if self.export_raw_data is True:
            # Backup old sensor config and set a new one with sensors everywhere
            old_sensor_config = scada_data.sensor_config
            scada_data.change_sensor_config(self.create_global_sensor_config(scada_data))

        sensor_readings = scada_data.get_data()
        sensor_readings_time = scada_data.sensor_readings_time
        col_desc = self.create_column_desc(scada_data)
        sensors_name = np.array([f"Sensor {i}" for i in range(1, sensor_readings.shape[1] + 1)],
                                dtype=object).reshape(-1, 1)
        col_desc = np.concatenate((sensors_name, col_desc), axis=1)

        if self.export_raw_data is True:
            # Restore old sensor config
            scada_data.change_sensor_config(old_sensor_config)

        with pd.ExcelWriter(self.f_out) as writer:
            pd.DataFrame(sensor_readings, columns=[f"Sensor {i}" for i in
                                                   range(1, sensor_readings.shape[1] + 1)]). \
                to_excel(writer, sheet_name="Sensor readings", index=False)
            pd.DataFrame(sensor_readings_time, columns=["Time (s)"]). \
                to_excel(writer, sheet_name="Sensor readings time", index=False)
            pd.DataFrame(col_desc, columns=["Name", "Type", "Location", "Unit"]). \
                to_excel(writer, sheet_name="Sensors description", index=False)


class ScadaDataMatlabExport(ScadaDataExport):
    """
    Class for exporting SCADA data to MATLAB (.mat file).
    """

    def export(self, scada_data: ScadaData) -> None:
        """
        Exports given SCADA data.

        Parameters
        ----------
        scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            SCADA data to be exported.
        """
        if not isinstance(scada_data, ScadaData):
            raise TypeError("'scada_data' must be an instance of " +
                            "'epyt_flow.simulation.scada_data.ScadaData' and not of " +
                            f"'{type(scada_data)}'")

        old_sensor_config = None
        if self.export_raw_data is True:
            # Backup old sensor config and set a new one with sensors everywhere
            old_sensor_config = scada_data.sensor_config
            scada_data.change_sensor_config(self.create_global_sensor_config(scada_data))

        sensor_readings = scada_data.get_data()
        sensor_readings_time = scada_data.sensor_readings_time
        col_desc = self.create_column_desc(scada_data)

        if self.export_raw_data is True:
            # Restore old sensor config
            scada_data.change_sensor_config(old_sensor_config)

        savemat(self.f_out, {"sensor_readings": sensor_readings,
                             "sensor_readings_time": sensor_readings_time,
                             "col_desc": col_desc,
                             "flow_unit": scada_data.sensor_config.flow_unit})



"""
Module provides a class for implementing model uncertainty.
"""
from typing import Optional
from copy import deepcopy
import warnings
import epyt
from epyt.epanet import ToolkitConstants
import numpy as np

from ..serialization import serializable, JsonSerializable, MODEL_UNCERTAINTY_ID
from .uncertainties import Uncertainty


@serializable(MODEL_UNCERTAINTY_ID, ".epytflow_uncertainty_model_uncertainty")
class ModelUncertainty(JsonSerializable):
    """
    Class implementing model uncertainty -- i.e. uncertainties in pipe length, pipe roughness,
    base demand, etc.

    Parameters
    ----------
    global_pipe_length_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of pipe lengths. None, in the case of no uncertainty.

        The default is None.
    global_pipe_roughness_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of pipe roughness coefficients. None, in the case of no uncertainty.

        The default is None.
    global_pipe_diameter_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of pipe diameters. None, in the case of no uncertainty.

        The default is None.
    global_base_demand_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of base demands. None, in the case of no uncertainty.

        The default is None.
    global_demand_pattern_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of demand patterns. None, in the case of no uncertainty.

        The default is None.
    global_elevation_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of elevations. None, in the case of no uncertainty.

        The default is None.
    global_constants_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of MSX constants. None, in the case of no uncertainty.

        The default is None.
    global_parameters_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of MSX parameters. None, in the case of no uncertaint.

        The default is None.
    local_pipe_length_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of pipe lengths -- i.e. a dictionary of pipe IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_pipe_roughness_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of pipe roughness coefficients -- i.e. a dictionary of pipe IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_pipe_diameter_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of pipe diameters -- i.e. a dictionary of pipe IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_base_demand_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of base demands -- i.e. a dictionary of node IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_demand_pattern_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of demand patterns --
        i.e. a dictionary of demand pattern IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_elevation_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of elevations -- i.e. a dictionary of node IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_constants_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of MSX constants -- i.e. a dictionary of constant IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_parameters_uncertainty : dict[tuple[str, int, str] :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of MSX parameters -- i.e. a dictionary of
        (parameter ID, item type, item ID) and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_patterns_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of EPANET patterns -- i.e. a dictionary of pattern IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_msx_patterns_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of EPANET-MSX patterns -- i.e. a dictionary of MSX pattern IDs
        and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    seed : `int`, optional
        Seed for the random number generator.

        Thed default is None.
    cache_original : `bool`, optional
        If True, all original properties are cached before the uncertainties are applied.
        This is necessary if you have multiple simulation runs and you want to reset/re-apply
        the uncertainties before each run.

        You can set it to False if you do not and want to re-apply the uncertainties and
        save some working memory.

        The default is True.
    """
    def __init__(self, global_pipe_length_uncertainty: Optional[Uncertainty] = None,
                 global_pipe_roughness_uncertainty: Optional[Uncertainty] = None,
                 global_pipe_diameter_uncertainty: Optional[Uncertainty] = None,
                 global_base_demand_uncertainty: Optional[Uncertainty] = None,
                 global_demand_pattern_uncertainty: Optional[Uncertainty] = None,
                 global_elevation_uncertainty: Optional[Uncertainty] = None,
                 global_constants_uncertainty: Optional[Uncertainty] = None,
                 global_parameters_uncertainty: Optional[Uncertainty] = None,
                 local_pipe_length_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_pipe_roughness_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_pipe_diameter_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_base_demand_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_demand_pattern_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_elevation_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_constants_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_parameters_uncertainty: Optional[dict[str, int, Uncertainty]] = None,
                 local_patterns_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_msx_patterns_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 seed: Optional[int] = None, cache_original: Optional[bool] = True,
                 **kwds):
        if global_pipe_length_uncertainty is not None:
            if not isinstance(global_pipe_length_uncertainty, Uncertainty):
                raise TypeError("'global_pipe_length_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_pipe_length_uncertainty)}'")
        if global_pipe_roughness_uncertainty is not None:
            if not isinstance(global_pipe_roughness_uncertainty, Uncertainty):
                raise TypeError("'global_pipe_roughness_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_pipe_roughness_uncertainty)}'")
        if global_pipe_diameter_uncertainty is not None:
            if not isinstance(global_pipe_diameter_uncertainty, Uncertainty):
                raise TypeError("'global_pipe_diameter_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_pipe_diameter_uncertainty)}'")
        if global_base_demand_uncertainty is not None:
            if not isinstance(global_base_demand_uncertainty, Uncertainty):
                raise TypeError("'global_base_demand_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_base_demand_uncertainty)}'")
        if global_demand_pattern_uncertainty is not None:
            if not isinstance(global_demand_pattern_uncertainty, Uncertainty):
                raise TypeError("'global_demand_pattern_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_demand_pattern_uncertainty)}'")
        if global_elevation_uncertainty is not None:
            if not isinstance(global_elevation_uncertainty, Uncertainty):
                raise TypeError("'global_elevation_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_elevation_uncertainty)}'")
        if global_constants_uncertainty is not None:
            if not isinstance(global_constants_uncertainty, Uncertainty):
                raise TypeError("'global_constants_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_constants_uncertainty)}'")
        if global_parameters_uncertainty is not None:
            if not isinstance(global_parameters_uncertainty, Uncertainty):
                raise TypeError("'global_parameters_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_parameters_uncertainty)}'")

        if local_pipe_length_uncertainty is not None:
            if not isinstance(local_pipe_length_uncertainty, dict):
                raise TypeError("'local_pipe_length_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_pipe_length_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_pipe_length_uncertainty.items()):
                raise TypeError("'local_pipe_length_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_pipe_roughness_uncertainty is not None:
            if not isinstance(local_pipe_roughness_uncertainty, dict):
                raise TypeError("'local_pipe_roughness_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_pipe_roughness_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_pipe_roughness_uncertainty.items()):
                raise TypeError("'local_pipe_roughness_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_pipe_diameter_uncertainty is not None:
            if not isinstance(local_pipe_diameter_uncertainty, dict):
                raise TypeError("'local_pipe_diameter_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_pipe_diameter_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_pipe_diameter_uncertainty.items()):
                raise TypeError("'local_pipe_diameter_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_base_demand_uncertainty is not None:
            if not isinstance(local_base_demand_uncertainty, dict):
                raise TypeError("'local_base_demand_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_base_demand_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_base_demand_uncertainty.items()):
                raise TypeError("'local_base_demand_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_demand_pattern_uncertainty is not None:
            if not isinstance(local_demand_pattern_uncertainty, dict):
                raise TypeError("'local_demand_pattern_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_demand_pattern_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_demand_pattern_uncertainty.items()):
                raise TypeError("'local_demand_pattern_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_elevation_uncertainty is not None:
            if not isinstance(local_elevation_uncertainty, dict):
                raise TypeError("'local_elevation_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_elevation_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_elevation_uncertainty.items()):
                raise TypeError("'local_elevation_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_constants_uncertainty is not None:
            if not isinstance(local_constants_uncertainty, dict):
                raise TypeError("'local_constants_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_constants_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_constants_uncertainty.items()):
                raise TypeError("'local_constants_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_parameters_uncertainty is not None:
            if not isinstance(local_parameters_uncertainty, dict):
                raise TypeError("'local_parameters_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_parameters_uncertainty)}'")
            if any(not isinstance(key, tuple) or not isinstance(key[0], str) or
                   not isinstance(key[1], int) or not isinstance(key[2], str) or
                   not isinstance(local_parameters_uncertainty[key], Uncertainty)
                   for key in local_parameters_uncertainty.keys()):
                raise TypeError("'local_parameters_uncertainty': " +
                                "All keys must be instances of 'tuple[str, int, str]' and all " +
                                "values must be instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_patterns_uncertainty is not None:
            if not isinstance(local_patterns_uncertainty, dict):
                raise TypeError("'local_patterns_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_patterns_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_patterns_uncertainty.items()):
                raise TypeError("'local_patterns_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_msx_patterns_uncertainty is not None:
            if not isinstance(local_msx_patterns_uncertainty, dict):
                raise TypeError("'local_msx_patterns_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_msx_patterns_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_msx_patterns_uncertainty.items()):
                raise TypeError("'local_msx_patterns_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if not isinstance(cache_original, bool):
            raise TypeError("'cache_original' must be an instance of 'bool' " +
                            f"but not of '{type(cache_original)}'")

        self._global_pipe_length = global_pipe_length_uncertainty
        self._global_pipe_roughness = global_pipe_roughness_uncertainty
        self._global_pipe_diameter = global_pipe_diameter_uncertainty
        self._global_base_demand = global_base_demand_uncertainty
        self._global_demand_pattern = global_demand_pattern_uncertainty
        self._global_elevation = global_elevation_uncertainty
        self._global_constants = global_constants_uncertainty
        self._global_parameters = global_parameters_uncertainty
        self._local_pipe_length = local_pipe_length_uncertainty
        self._local_pipe_roughness = local_pipe_roughness_uncertainty
        self._local_pipe_diameter = local_pipe_diameter_uncertainty
        self._local_base_demand = local_base_demand_uncertainty
        self._local_demand_pattern = local_demand_pattern_uncertainty
        self._local_elevation = local_elevation_uncertainty
        self._local_constants = local_constants_uncertainty
        self._local_parameters = local_parameters_uncertainty
        self._local_patterns = local_patterns_uncertainty
        self._local_msx_patterns = local_msx_patterns_uncertainty
        self.__seed = seed
        self.__cache_original = cache_original

        self._cache_links_length = None
        self._cache_links_diameter = None
        self._cache_links_roughness_coeff = None
        self._cache_nodes_base_demand = None
        self._cache_nodes_demand_pattern = None
        self._cache_nodes_elevation = None
        self._cache_patterns = None
        self._cache_msx_constants = None
        self._cache_msx_links_parameters = None
        self._cache_msx_tanks_parameters = None
        self._cache_msx_patterns = None

        super().__init__(**kwds)

    @property
    def seed(self) -> int:
        """
        Returns the seed used for the random number generator.

        Returns
        -------
        `int`
            Seed for the random number generator.
        """
        return self.__seed

    @property
    def global_pipe_length(self) -> Uncertainty:
        """
        Returns the global pipe length uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global pipe length uncertainty.
        """
        return deepcopy(self._global_pipe_length)

    @property
    def global_pipe_roughness(self) -> Uncertainty:
        """
        Returns the global pipe roughness uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global pipe roughness uncertainty.
        """
        return deepcopy(self._global_pipe_roughness)

    @property
    def global_pipe_diameter(self) -> Uncertainty:
        """
        Returns the global pipe diameter uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global pipe diameter uncertainty.
        """
        return deepcopy(self._global_pipe_diameter)

    @property
    def global_base_demand(self) -> Uncertainty:
        """
        Returns the global base demand uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global base demand uncertainty.
        """
        return deepcopy(self._global_base_demand)

    @property
    def global_demand_pattern(self) -> Uncertainty:
        """
        Returns the global demand pattern uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global demand pattern uncertainty.
        """
        return deepcopy(self._global_demand_pattern)

    @property
    def global_elevation(self) -> Uncertainty:
        """
        Returns the global node elevation uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global node elevation uncertainty.
        """
        return deepcopy(self._global_elevation)

    @property
    def global_constants(self) -> Uncertainty:
        """
        Returns the global MSX constant uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global MSX constant uncertainty.
        """
        return deepcopy(self._global_constants)

    @property
    def global_parameters(self) -> Uncertainty:
        """
        Returns the global MSX parameter uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global MSX parameter uncertainty.
        """
        return deepcopy(self._global_parameters)

    @property
    def local_pipe_length(self) -> dict[str, Uncertainty]:
        """
        Returns the local pipe length uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local pipe length uncertainty.
        """
        return deepcopy(self._local_pipe_length)

    @property
    def local_pipe_roughness(self) -> dict[str, Uncertainty]:
        """
        Returns the local pipe roughness uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local pipe roughness uncertainty.
        """
        return deepcopy(self._local_pipe_roughness)

    @property
    def local_pipe_diameter(self) -> dict[str, Uncertainty]:
        """
        Returns the local pipe diameter uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local pipe diameter uncertainty.
        """
        return deepcopy(self._local_pipe_diameter)

    @property
    def local_base_demand(self) -> dict[str, Uncertainty]:
        """
        Returns the local base demand uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local base demand uncertainty.
        """
        return deepcopy(self._local_base_demand)

    @property
    def local_demand_pattern(self) -> dict[str, Uncertainty]:
        """
        Returns the local demand pattern uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local demand pattern uncertainty.
        """
        return deepcopy(self._local_demand_pattern)

    @property
    def local_elevation(self) -> dict[str, Uncertainty]:
        """
        Returns the local node elevation uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local node elevation uncertainty.
        """
        return deepcopy(self._local_elevation)

    @property
    def local_constants(self) -> dict[str, Uncertainty]:
        """
        Returns the local MSX constant uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local MSX constant uncertainty.
        """
        return deepcopy(self._local_constants)

    @property
    def local_parameters(self) -> dict[tuple[str, int, str], Uncertainty]:
        """
        Returns the local MSX parameter uncertainty.

        Returns
        -------
        dict[tuple[str, int, str], :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local MSX parameter uncertainty.
        """
        return deepcopy(self._local_parameters)

    @property
    def local_patterns(self) -> dict[str, Uncertainty]:
        """
        Returns the local EPANET patterns uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local EPANET patterns uncertainty.
        """
        return deepcopy(self._local_patterns)

    @property
    def local_msx_patterns(self) -> dict[str, Uncertainty]:
        """
        Returns the local EPANET-MSX patterns uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local EPANET-MSX patterns uncertainty.
        """
        return deepcopy(self._local_msx_patterns)

    def get_attributes(self) -> dict:
        attribs = {"global_pipe_length_uncertainty": self._global_pipe_length,
                   "global_pipe_roughness_uncertainty": self._global_pipe_roughness,
                   "global_pipe_diameter_uncertainty": self._global_pipe_diameter,
                   "global_base_demand_uncertainty": self._global_base_demand,
                   "global_demand_pattern_uncertainty": self._global_demand_pattern,
                   "global_elevation_uncertainty": self._global_elevation,
                   "global_constants_uncertainty": self._global_constants,
                   "global_parameters_uncertainty": self._global_parameters,
                   "local_pipe_length_uncertainty": self._local_pipe_length,
                   "local_pipe_roughness_uncertainty": self._local_pipe_roughness,
                   "local_pipe_diameter_uncertainty": self._local_pipe_diameter,
                   "local_base_demand_uncertainty": self._local_base_demand,
                   "local_demand_pattern_uncertainty": self._local_demand_pattern,
                   "local_elevation_uncertainty": self._local_elevation,
                   "local_constants_uncertainty": self._local_constants,
                   "local_parameters_uncertainty": self._local_parameters,
                   "local_patterns_uncertainty": self._local_patterns,
                   "local_msx_patterns_uncertainty": self._local_msx_patterns,
                   "seed": self.__seed}

        return super().get_attributes() | attribs

    def __eq__(self, other) -> bool:
        if not isinstance(other, ModelUncertainty):
            raise TypeError("Can not compare 'ModelUncertainty' instance " +
                            f"with '{type(other)}' instance")

        return self._global_pipe_length == other.global_pipe_length \
            and self._global_pipe_roughness == other.global_pipe_roughness \
            and self._global_pipe_diameter == other.global_pipe_diameter \
            and self._global_base_demand == other.global_base_demand \
            and self._global_demand_pattern == other.global_demand_pattern \
            and self._global_elevation == other.global_elevation \
            and self._global_parameters == other.global_parameters \
            and self._global_constants == other.global_constants \
            and self._local_pipe_length == other.local_pipe_length \
            and self._local_pipe_roughness == other.local_pipe_roughness \
            and self._local_pipe_diameter == other.local_pipe_diameter \
            and self._local_base_demand == other.local_base_demand \
            and self._local_demand_pattern == other.local_demand_pattern \
            and self._local_elevation == other.local_elevation \
            and self._local_parameters == other.local_parameters \
            and self._local_constants == other.local_constants \
            and self._local_patterns == other.local_patterns \
            and self._local_msx_patterns == other.local_msx_patterns \
            and self.__seed == other.seed

    def __str__(self) -> str:
        return f"global_pipe_length: {self._global_pipe_length} " +\
            f"global_pipe_roughness: {self._global_pipe_roughness} " + \
            f"global_pipe_diameter: {self._global_pipe_diameter} " + \
            f"global_demand_base: {self._global_base_demand} " + \
            f"global_demand_pattern: {self._global_demand_pattern} " + \
            f"global_elevation: {self._global_elevation} " + \
            f"global_constants: {self._global_constants} " + \
            f"global_parameters: {self._global_parameters}" + \
            f"local_pipe_length: {self._local_pipe_length} " +\
            f"local_pipe_roughness: {self._local_pipe_roughness} " + \
            f"local_pipe_diameter: {self._local_pipe_diameter} " + \
            f"local_demand_base: {self._local_base_demand} " + \
            f"local_demand_pattern: {self._local_demand_pattern} " + \
            f"local_elevation: {self._local_elevation} " + \
            f"local_constants: {self._local_constants} " + \
            f"local_parameters: {self._local_parameters} " + \
            f"local_patterns: {self._local_patterns} " + \
            f"local_msx_patterns: {self._local_msx_patterns} + seed: {self.__seed}"

    def undo(self, epanet_api: epyt.epanet) -> None:
        """
        Undo all applied uncertainties -- i.e, resets the properties to their original value.

        Note that this function can only be used if `cache_original` (of the constructor)
        was set to True (default).

        Parameters
        ----------
        epanet_api : `epyt.epanet <https://epanet-python-toolkit-epyt.readthedocs.io/en/stable/api.html#epyt.epanet.epanet>`_
            Interface to EPANET and EPANET-MSX
        """
        if self.__cache_original is False:
            raise ValueError("Caching was disabled by the user")

        if self._cache_links_length is not None:
            epanet_api.setLinkLength(self._cache_links_length)

        if self._cache_links_diameter is not None:
            epanet_api.setLinkDiameter(self._cache_links_diameter)

        if self._cache_links_roughness_coeff is not None:
            epanet_api.setLinkRoughnessCoeff(self._cache_links_roughness_coeff)

        if self._cache_nodes_base_demand is not None:
            for node_idx in self._cache_nodes_base_demand.keys():
                for demand_category, base_demand in self._cache_nodes_base_demand[node_idx].items():
                    epanet_api.setNodeBaseDemands(node_idx, demand_category + 1, base_demand)

        if self._cache_nodes_demand_pattern is not None:
            for pattern_id, demand_pattern in self._cache_nodes_demand_pattern.items():
                for t, v in enumerate(demand_pattern):
                    epanet_api.setPatternValue(pattern_id, t+1, v)

        if self._cache_nodes_elevation is not None:
            epanet_api.setNodeElevations(self._cache_nodes_elevation)

        if self._cache_patterns is not None:
            for pattern_idx, pattern in self._cache_patterns.items():
                epanet_api.setPattern(pattern_idx, pattern)

        if self._cache_msx_constants is not None:
            epanet_api.setMSXConstantsValue(self._cache_msx_constants)

        if self._cache_msx_links_parameters is not None:
            for pipe_idx, parameters_pipes_val in self._cache_msx_links_parameters.items():
                epanet_api.setMSXParametersPipesValue(pipe_idx, parameters_pipes_val)

        if self._cache_msx_tanks_parameters is not None:
            for tank_idx, parameters_tanks_val in self._cache_msx_tanks_parameters.items():
                epanet_api.setMSXParametersTanksValue(tank_idx, parameters_tanks_val)

        if self._cache_msx_patterns is not None:
            for pattern_idx, pattern in self._cache_msx_patterns:
                epanet_api.setMSXPattern(pattern_idx, pattern)

    def apply(self, epanet_api: epyt.epanet) -> None:
        """
        Applies the specified model uncertainties to the scenario.

        Parameters
        ----------
        epanet_api : `epyt.epanet <https://epanet-python-toolkit-epyt.readthedocs.io/en/stable/api.html#epyt.epanet.epanet>`_
            Interface to EPANET and EPANET-MSX.
        """
        np_rand_gen = np.random.default_rng(seed=self.__seed)

        if self._global_pipe_length is not None:
            self._global_pipe_length.set_random_generator(np_rand_gen)

            link_length = epanet_api.getLinkLength()
            self._cache_links_length = np.copy(link_length)

            link_length = self._global_pipe_length.apply_batch(link_length)
            epanet_api.setLinkLength(link_length)

        if self._local_pipe_length is not None:
            self._local_pipe_length.set_random_generator(np_rand_gen)
            self._cache_links_length = epanet_api.getLinkLength()

            for pipe_id, uncertainty in self._local_pipe_length.items():
                link_idx = epanet_api.getLinkIndex(pipe_id)
                link_length = epanet_api.getLinkLength(link_idx)

                link_length = uncertainty.apply(link_length)
                epanet_api.setLinkLength(link_idx, link_length)

        if self._global_pipe_diameter is not None:
            self._global_pipe_diameter.set_random_generator(np_rand_gen)

            link_diameters = epanet_api.getLinkDiameter()
            self._cache_links_diameter = np.copy(link_diameters)

            link_diameters = self._global_pipe_diameter.apply_batch(link_diameters)
            epanet_api.setLinkDiameter(link_diameters)

        if self._local_pipe_diameter is not None:
            self._local_pipe_diameter.set_random_generator(np_rand_gen)
            self._cache_links_diameter = epanet_api.getLinkDiameter()

            for pipe_id, uncertainty in self._local_pipe_diameter.items():
                link_idx = epanet_api.getLinkIndex(pipe_id)
                link_diameter = epanet_api.getLinkDiameter(link_idx)

                link_diameter = uncertainty.apply(link_diameter)
                epanet_api.setLinkDiameter(link_idx, link_diameter)

        if self._global_pipe_roughness is not None:
            self._global_pipe_roughness.set_random_generator(np_rand_gen)

            coeffs = epanet_api.getLinkRoughnessCoeff()
            self._cache_links_roughness_coeff = np.copy(coeffs)

            coeffs = self._global_pipe_roughness.apply_batch(coeffs)
            epanet_api.setLinkRoughnessCoeff(coeffs)

        if self._local_pipe_roughness is not None:
            self._local_pipe_roughness.set_random_generator(np_rand_gen)
            self._cache_links_roughness_coeff = epanet_api.getLinkRoughnessCoeff()

            for pipe_id, uncertainty in self._local_pipe_roughness.items():
                link_idx = epanet_api.getLinkIndex(pipe_id)
                link_roughness_coeff = epanet_api.getLinkRoughnessCoeff(link_idx)

                link_roughness_coeff = uncertainty.apply(link_roughness_coeff)
                epanet_api.setLinkRoughnessCoeff(link_idx, link_roughness_coeff)

        if self._global_base_demand is not None:
            self._global_base_demand.set_random_generator(np_rand_gen)

            self._cache_nodes_base_demand = {}
            all_nodes_idx = epanet_api.getNodeIndex()
            for node_idx in all_nodes_idx:
                self._cache_nodes_base_demand[node_idx] = {}
                n_demand_categories = epanet_api.getNodeDemandCategoriesNumber(node_idx)
                for demand_category in range(n_demand_categories):
                    base_demand = epanet_api.getNodeBaseDemands(node_idx)[demand_category + 1]
                    self._cache_nodes_base_demand[node_idx][demand_category] = base_demand

                    base_demand = self._global_base_demand.apply(base_demand)
                    epanet_api.setNodeBaseDemands(node_idx, demand_category + 1, base_demand)

        if self._local_base_demand is not None:
            self._local_base_demand.set_random_generator(np_rand_gen)

            self._cache_nodes_base_demand = {}
            for node_id, uncertainty in self._local_base_demand.items():
                node_idx = epanet_api.getNodeIndex(node_id)
                self._cache_nodes_base_demand[node_idx] = {}
                n_demand_categories = epanet_api.getNodeDemandCategoriesNumber(node_idx)
                for demand_category in range(n_demand_categories):
                    base_demand = epanet_api.getNodeBaseDemands(node_idx)[demand_category + 1]
                    self._cache_nodes_base_demand[node_idx][demand_category] = base_demand

                    base_demand = uncertainty.apply(base_demand)
                    epanet_api.setNodeBaseDemands(node_idx, demand_category + 1, base_demand)

        if self._global_demand_pattern is not None:
            self._global_demand_pattern.set_random_generator(np_rand_gen)

            self._cache_nodes_demand_pattern = {}
            demand_patterns_idx = epanet_api.getNodeDemandPatternIndex()
            demand_patterns_id = np.unique([demand_patterns_idx[k]
                                            for k in demand_patterns_idx.keys()])
            for pattern_id in demand_patterns_id:
                if pattern_id == 0:
                    continue

                pattern_length = epanet_api.getPatternLengths(pattern_id)
                demand_pattern = np.zeros(pattern_length)

                for t in range(pattern_length):
                    v = epanet_api.getPatternValue(pattern_id, t+1)
                    demand_pattern[t] = v

                    v_ = self._global_demand_pattern.apply(v)
                    epanet_api.setPatternValue(pattern_id, t+1, v_)

                self._cache_nodes_demand_pattern[pattern_id] = demand_pattern

        if self._local_demand_pattern is not None:
            self._local_demand_pattern.set_random_generator(np_rand_gen)

            self._cache_nodes_demand_pattern = {}
            patterns_id = epanet_api.getPatternNameID()
            paterns_idx = epanet_api.getPatternIndex()

            for pattern_id, uncertainty in self._local_demand_pattern.items():
                pattern_idx = paterns_idx[patterns_id.index(pattern_id)]
                pattern_length, = epanet_api.getPatternLengths(pattern_id)
                demand_pattern = np.zeros(pattern_length)

                for t in range(pattern_length):
                    v = epanet_api.getPatternValue(pattern_idx, t+1)
                    demand_pattern[t] = v

                    v_ = uncertainty.apply(v)
                    epanet_api.setPatternValue(pattern_idx, t+1, v_)

                self._cache_nodes_demand_pattern[pattern_id] = demand_pattern

        if self._global_elevation is not None:
            self._global_elevation.set_random_generator(np_rand_gen)

            elevations = epanet_api.getNodeElevations()
            self._cache_nodes_elevation = np.copy(elevations)

            elevations = self._global_elevation.apply_batch(elevations)
            epanet_api.setNodeElevations(elevations)

        if self._local_elevation is not None:
            self._local_elevation.set_random_generator(np_rand_gen)
            self._cache_nodes_elevation = epanet_api.getNodeElevations()

            for node_id, uncertainty in self._local_elevation.items():
                node_idx = epanet_api.getNodeIndex(node_id)
                elevation = epanet_api.getNodeElevations(node_idx)

                elevation = uncertainty.apply(elevation)
                epanet_api.setNodeElevations(node_idx, elevation)

        if self._local_patterns is not None:
            self._local_patterns.set_random_generator(np_rand_gen)
            self._cache_patterns = {}

            for pattern_id, uncertainty in self._local_patterns.items():
                pattern_idx = epanet_api.getPatternIndex(pattern_id)
                pattern_length = epanet_api.getPatternLengths(pattern_idx)
                pattern = np.array([epanet_api.getPatternValue(pattern_idx, t+1)
                                    for t in range(pattern_length)])
                self._cache_patterns[pattern_idx] = pattern

                pattern = uncertainty.apply_batch(pattern)
                epanet_api.setPattern(pattern_idx, pattern)

        if epanet_api.MSXFile is not None:
            if self._global_constants is not None:
                self._global_constants.set_random_generator(np_rand_gen)

                constants = np.array(epanet_api.getMSXConstantsValue())
                self._cache_msx_patterns = np.copy(constants)

                constants = self._global_constants.apply_batch(constants)
                epanet_api.setMSXConstantsValue(constants)

            if self._local_constants:
                self._local_constants.set_random_generator(np_rand_gen)

                self._cache_msx_patterns = np.array(epanet_api.getMSXConstantsValue())

                for constant_id, uncertainty in self._local_constants.items():
                    idx = epanet_api.MSXgetindex(ToolkitConstants.MSX_CONSTANT, constant_id)
                    constant = epanet_api.msx.MSXgetconstant(idx)

                    constant = uncertainty.apply(constant)
                    epanet_api.msx.MSXsetconstant(idx, constant)

            if self._global_parameters is not None:
                self._global_parameters.set_random_generator(np_rand_gen)

                self._cache_msx_links_parameters = {}
                parameters_pipes = epanet_api.getMSXParametersPipesValue()
                for i, pipe_idx in enumerate(epanet_api.getLinkPipeIndex()):
                    if len(parameters_pipes[i]) == 0:
                        continue

                    self._cache_msx_links_parameters[pipe_idx] = np.array(parameters_pipes[i])
                    parameters_pipes_val = self._global_parameters.apply_batch(
                        np.array(parameters_pipes[i]))
                    epanet_api.setMSXParametersPipesValue(pipe_idx, parameters_pipes_val)

                self._cache_msx_tanks_parameters = {}
                parameters_tanks = epanet_api.getMSXParametersTanksValue()
                for i, tank_idx in enumerate(epanet_api.getNodeTankIndex()):
                    if parameters_tanks[i] is None or len(parameters_tanks[i]) == 0:
                        continue

                    self._cache_msx_tanks_parameters[tank_idx] = np.array(parameters_tanks[i])
                    parameters_tanks_val = self._global_parameters.apply_batch(
                        np.array(parameters_tanks[i]))
                    epanet_api.setMSXParametersTanksValue(tank_idx, parameters_tanks_val)

            if self._local_parameters is not None:
                self._local_parameters.set_random_generator(np_rand_gen)
                self._cache_msx_links_parameters = {}
                self._cache_msx_tanks_parameters = {}

                for (param_id, item_type, item_id), uncertainty in self._local_parameters.items():
                    idx, = epanet_api.getMSXParametersIndex([param_id])

                    if item_type == ToolkitConstants.MSX_NODE:
                        item_idx = epanet_api.getNodeIndex(item_id)
                    elif item_type == ToolkitConstants.MSX_LINK:
                        item_idx = epanet_api.getLinkIndex(item_id)
                    else:
                        raise ValueError(f"Unknown item type '{item_type}' must be either " +
                                         "ToolkitConstants.MSX_NODE or ToolkitConstants.MSX_LINK")

                    parameter = epanet_api.msx.MSXgetparameter(item_type, item_idx, idx)
                    if item_type == ToolkitConstants.MSX_NODE:
                        self._cache_msx_tanks_parameters[item_idx] = parameter
                    elif item_type == ToolkitConstants.MSX_LINK:
                        self._cache_msx_links_parameters[item_idx] = parameter

                    parameter = uncertainty.apply(parameter)
                    epanet_api.msx.MSXsetparameter(item_type, item_idx, idx, parameter)

            if self._local_msx_patterns is not None:
                self._local_msx_patterns.set_random_generator(np_rand_gen)
                self._cache_msx_patterns = {}

                for pattern_id, uncertainty in self._local_msx_patterns.items():
                    pattern_idx, = epanet_api.getMSXPatternsIndex([pattern_id])
                    pattern = epanet_api.getMSXConstantsValue([pattern_idx])
                    self._cache_msx_patterns[pattern_idx] = np.copy(pattern)

                    pattern = uncertainty.apply_batch(pattern)
                    epanet_api.setMSXPattern(pattern_idx, pattern)
        else:
            if self._local_msx_patterns is not None or self._local_parameters is not None or \
                    self._local_constants is not None or self._global_constants is not None or \
                    self._global_parameters is not None:
                warnings.warn("Ignoring EPANET-MSX uncertainties because not .msx file was loaded")



"""
Module provides helper functions.
"""
import os
import math
import tempfile
import zipfile
from pathlib import Path
import requests
from tqdm import tqdm
import numpy as np
import matplotlib
import matplotlib.pyplot as plt


def time_points_to_one_hot_encoding(time_points: list[int], total_length: int) -> list[int]:
    """
    Converts a list of time points into a one-hot-encoding.

    Parameters
    ----------
    time_points : `list[int]`
        Time points to be one-hot-encoded.
    total_length : `int`
        Length of final one-hot-encoding.

    Returns
    -------
    `list[int]`
        One-hot-encoded time points.
    """
    results = [0] * total_length

    for t in time_points:
        results[t] = 1

    return results


def volume_to_level(tank_volume: float, tank_diameter: float) -> float:
    """
    Computes the water level in a tank containing a given volume of water.

    Parameters
    ----------
    tank_volume : `float`
        Water volume in the tank.
    tank_diameter : `float`
        Diameter of the tank.

    Returns
    -------
    `float`
        Water level in tank.
    """
    if not isinstance(tank_volume, float):
        raise TypeError("'tank_volume' must be an instace of 'float' " +
                        f"but not of '{type(tank_volume)}'")
    if tank_volume < 0:
        raise ValueError("'tank_volume' can not be negative")
    if not isinstance(tank_diameter, float):
        raise TypeError("'tank_diameter' must be an instace of 'float' " +
                        f"but not of '{type(tank_diameter)}'")
    if tank_diameter <= 0:
        raise ValueError("'tank_diameter' must be greater than zero")

    return (4. / (math.pow(tank_diameter, 2) * math.pi)) * tank_volume


def level_to_volume(tank_level: float, tank_diameter: float) -> float:
    """
    Computes the volume of water in a tank given the water level in the tank.

    Parameters
    ----------
    tank_level : `float`
        Water level in the tank.
    tank_diameter : `float`
        Diameter of the tank.

    Returns
    -------
    `float`
        Water volume in tank.
    """
    if not isinstance(tank_level, float):
        raise TypeError("'tank_level' must be an instace of 'float' " +
                        f"but not of '{type(tank_level)}'")
    if tank_level < 0:
        raise ValueError("'tank_level' can not be negative")
    if not isinstance(tank_diameter, float):
        raise TypeError("'tank_diameter' must be an instace of 'float' " +
                        f"but not of '{type(tank_diameter)}'")
    if tank_diameter <= 0:
        raise ValueError("'tank_diameter' must be greater than zero")

    return tank_level * math.pow(0.5 * tank_diameter, 2) * math.pi


def plot_timeseries_data(data: np.ndarray, labels: list[str] = None, x_axis_label: str = None,
                         y_axis_label: str = None, y_ticks: tuple[list[float], list[str]] = None,
                         show: bool = True, save_to_file: str = None,
                         ax: matplotlib.axes.Axes = None) -> matplotlib.axes.Axes:
    """
    Plots a single or multiple time series.

    Parameters
    ----------
    data : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Time series data -- each row in `data` corresponds to a complete time series.
    labels : `list[str]`, optional
        Labels for each time series in `data`.
        If None, no labels are shown.

        The default is None.
    x_axis_label : `str`, optional
        X axis label.

        The default is None.
    y_axis_label : `str`, optional
        Y axis label.

        The default is None.
    y_ticks: `(list[float], list[str])`, optional
        Tuple of ticks (numbers) and labels (strings) for the y-axis.

        The default is None.
    show : `bool`, optional
        If True, the plot/figure is shown in a window.

        Only considered when 'ax' is None.

        The default is True.
    save_to_file : `str`, optional
        File to which the plot is saved.

        If specified, 'show' must be set to False --
        i.e. a plot can not be shown and saved to a file at the same time!

        The default is None.
    ax : `matplotlib.axes.Axes <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.html>`_, optional
        If not None, 'ax' is used for plotting.

        The default is None.

    Returns
    -------
    `matplotlib.axes.Axes <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.html>`_
        Plot.
    """
    if not isinstance(data, np.ndarray):
        raise TypeError(f"'data' must be an instance of 'numpy.ndarray' but not of '{type(data)}'")
    if len(data.shape) != 2:
        raise ValueError("'data' must be a 2d array where each row corresponds to a time series " +
                         "-- use '.reshape(1, -1)' in case of single time series")
    if labels is not None:
        if not isinstance(labels, list) or not all(isinstance(label, str) for label in labels):
            raise TypeError("'labels' must be a instance of 'list[str]'")
    if x_axis_label is not None:
        if not isinstance(x_axis_label, str):
            raise TypeError("'x_axis_label' must be an instance of 'str' " +
                            f"but not of '{type(x_axis_label)}'")
    if y_axis_label is not None:
        if not isinstance(y_axis_label, str):
            raise TypeError("'y_axis_label' must be an instance of 'str' " +
                            f"but not of '{type(y_axis_label)}'")
    if y_ticks is not None:
        if len(y_ticks) != 2:
            raise ValueError("'y_ticks' must be a tuple ticks (numbers) and labels (strings)")
    if not isinstance(show, bool):
        raise TypeError(f"'show' must be an instance of 'bool' but not of '{type(show)}'")
    if save_to_file is not None:
        if show is True:
            raise ValueError("'show' must be False if 'save_to_file' is set")

        if not isinstance(save_to_file, str):
            raise TypeError("'save_to_file' must be an instance of 'str' but not of " +
                            f"'{type(save_to_file)}'")
    if ax is not None:
        if not isinstance(ax, matplotlib.axes.Axes):
            raise TypeError("ax' must be an instance of 'matplotlib.axes.Axes'" +
                            f"but not of '{type(ax)}'")

    fig = None
    if ax is None:
        fig, ax = plt.subplots()

    labels = labels if labels is not None else [None] * data.shape[0]

    for i in range(data.shape[0]):
        ax.plot(data[i, :], ".-", label=labels[i])

    if not any(label is None for label in labels):
        ax.legend()

    if x_axis_label is not None:
        ax.set_xlabel(x_axis_label)
    if y_axis_label is not None:
        ax.set_ylabel(y_axis_label)
    if y_ticks is not None:
        yticks_pos, yticks_labels = y_ticks
        ax.set_yticks(yticks_pos, labels=yticks_labels)

    if show is True and fig is not None:
        plt.show()
    if save_to_file is not None:
        folder_path = str(Path(save_to_file).parent.absolute())
        create_path_if_not_exist(folder_path)

        if fig is None:
            plt.savefig(save_to_file, bbox_inches='tight')
        else:
            fig.savefig(save_to_file, bbox_inches='tight')

    return ax


def plot_timeseries_prediction(y: np.ndarray, y_pred: np.ndarray,
                               confidence_interval: np.ndarray = None,
                               x_axis_label: str = None, y_axis_label: str = None,
                               y_ticks: tuple[list[float], list[str]] = None,
                               show: bool = True, save_to_file: str = None,
                               ax: matplotlib.axes.Axes = None
                               ) -> matplotlib.axes.Axes:
    """
    Plots the prediction (e.g. forecast) of *single* time series together with the
    ground truth time series. In addition, confidence intervals can be plotted as well.

    Parameters
    ----------
    y : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Ground truth values.
    y_pred : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Predicted values.
    confidence_interval : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, optional
        Confidence interval (upper and lower value) for each prediction in `y_pred`.
        If not None, the confidence interval is plotted as well.

        The default is None.
    x_axis_label : `str`, optional
        X axis label.

        The default is None.
    y_axis_label : `str`, optional
        Y axis label.

        The default is None.
    y_ticks: `(list[float], list[str])`, optional
        Tuple of ticks (numbers) and labels (strings) for the y-axis.

        The default is None.
    show : `bool`, optional
        If True, the plot/figure is shown in a window.

        Only considered when 'ax' is None.

        The default is True.
    save_to_file : `str`, optional
        File to which the plot is saved.

        If specified, 'show' must be set to False --
        i.e. a plot can not be shown and saved to a file at the same time!

        The default is None.
    ax : `matplotlib.axes.Axes <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.html>`_, optional
        If not None, 'axes' is used for plotting.

        The default is None.

    Returns
    -------
    `matplotlib.axes.Axes <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.html>`_
        Plot.
    """
    if not isinstance(y_pred, np.ndarray):
        raise TypeError("'y_pred' must be an instance of 'numpy.ndarray' " +
                        f"but not of '{type(y_pred)}'")
    if not isinstance(y, np.ndarray):
        raise TypeError("'y' must be an instance of 'numpy.ndarray' " +
                        f"but not of '{type(y)}'")
    if y_pred.shape != y.shape:
        raise ValueError(f"Shape mismatch: {y_pred.shape} vs. {y.shape}")
    if len(y_pred.shape) != 1:
        raise ValueError("'y_pred' must be a 1d array")
    if len(y.shape) != 1:
        raise ValueError("'y' must be a 1d array")
    if x_axis_label is not None:
        if not isinstance(x_axis_label, str):
            raise TypeError("'x_axis_label' must be an instance of 'str' " +
                            f"but not of '{type(x_axis_label)}'")
    if y_axis_label is not None:
        if not isinstance(y_axis_label, str):
            raise TypeError("'y_axis_label' must be an instance of 'str' " +
                            f"but not of '{type(y_axis_label)}'")
    if y_ticks is not None:
        if len(y_ticks) != 2:
            raise ValueError("'y_ticks' must be a tuple ticks (numbers) and labels (strings)")
    if not isinstance(show, bool):
        raise TypeError(f"'show' must be an instance of 'bool' but not of '{type(show)}'")
    if save_to_file is not None:
        if show is True:
            raise ValueError("'show' must be False if 'save_to_file' is set")

        if not isinstance(save_to_file, str):
            raise TypeError("'save_to_file' must be an instance of 'str' but not of " +
                            f"'{type(save_to_file)}'")
    if ax is not None:
        if not isinstance(ax, matplotlib.axes.Axes):
            raise TypeError("ax' must be an instance of 'matplotlib.axes.Axes'" +
                            f"but not of '{type(ax)}'")

    fig = None
    if ax is None:
        fig, ax = plt.subplots()

    if confidence_interval is not None:
        ax.fill_between(range(len(y_pred)),
                        y_pred - confidence_interval[0],
                        y_pred + confidence_interval[1],
                        alpha=0.5)
    ax.plot(y, ".-", label="Ground truth")
    ax.plot(y_pred, ".-", label="Prediction")
    ax.legend()

    if x_axis_label is not None:
        ax.set_xlabel(x_axis_label)
    if y_axis_label is not None:
        ax.set_ylabel(y_axis_label)
    if y_ticks is not None:
        yticks_pos, yticks_labels = y_ticks
        ax.set_yticks(yticks_pos, labels=yticks_labels)

    if show is True and fig is not None:
        plt.show()
    if save_to_file is not None:
        folder_path = str(Path(save_to_file).parent.absolute())
        create_path_if_not_exist(folder_path)

        if fig is None:
            plt.savefig(save_to_file, bbox_inches='tight')
        else:
            fig.savefig(save_to_file, bbox_inches='tight')

    return ax


def download_if_necessary(download_path: str, url: str, verbose: bool = True,
                          backup_urls: list[str] = []) -> None:
    """
    Downloads a file from a given URL if it does not already exist in a given path.

    Note that if the path (folder) does not already exist, it will be created.

    Parameters
    ----------
    download_path : `str`
        Local path to the file -- if this path does not exist, the file will be downloaded from
        the provided 'url' and stored in 'download_dir'.
    url : `str`
        Web-URL.
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    backup_urls : `list[str]`, optional
        List of alternative URLs that are being tried in the case that downloading from 'url' fails.

        The default is an empty list.
    """
    folder_path = str(Path(download_path).parent.absolute())
    create_path_if_not_exist(folder_path)

    if not os.path.isfile(download_path):
        response = requests.get(url, stream=verbose, allow_redirects=True, timeout=1000)

        if response.status_code != 200:
            for backup_url in backup_urls:
                response = requests.get(backup_url, stream=verbose, allow_redirects=True,
                                        timeout=1000)
                if response.status_code == 200:
                    break
        if response.status_code != 200:
            raise SystemError(f"Failed to download -- {response.status_code}")

        if verbose is True:
            content_length = int(response.headers.get('content-length', 0))
            with open(download_path, "wb") as file, tqdm(desc=download_path,
                                                         total=content_length,
                                                         ascii=True,
                                                         unit='B',
                                                         unit_scale=True,
                                                         unit_divisor=1024) as progress_bar:
                for data in response.iter_content(chunk_size=1024):
                    size = file.write(data)
                    progress_bar.update(size)
        else:
            with open(download_path, "wb") as f_out:
                f_out.write(response.content)


def create_path_if_not_exist(path_in: str) -> None:
    """
    Creates a directory and all its parent directories if they do not already exist.

    Parameters
    ----------
    path_in : `str`
        Path to be created.
    """
    Path(path_in).mkdir(parents=True, exist_ok=True)


def pack_zip_archive(f_in: list[str], f_out: str) -> None:
    """
    Compresses a given list of files into a .zip archive.

    Parameters
    ----------
    f_in : `list[str]`
        List of files to be compressed into the .zip archive.
    f_out : `str`
        Path to the final .zip file.
    """
    with zipfile.ZipFile(f_out, "w") as f_zip_out:
        for f_cur_in in f_in:
            f_zip_out.write(f_cur_in, compress_type=zipfile.ZIP_DEFLATED)


def unpack_zip_archive(f_in: str, folder_out: str) -> None:
    """
    Unpacks a .zip archive.

    Parameters
    ----------
    f_in : `str`
        Path to the .zip file.
    folder_out : `str`
        Path to the folder where the unpacked files will be stored.
    """
    with zipfile.ZipFile(f_in, "r") as f:
        f.extractall(folder_out)


def get_temp_folder() -> str:
    """
    Gets a path to a temporary folder -- i.e. a folder for storing temporary files.

    Returns
    -------
    `str`
        Path to a temporary folder.
    """
    return tempfile.gettempdir()


def to_seconds(days: int = None, hours: int = None, minutes: int = None) -> int:
    """
    Converts a timestamp (i.e. days, hours, minutes) into seconds.

    Parameters
    ----------
    days : `int`, optional
        Days.
    hours : `int`, optional
        Hours.
    minutes : `int`, optional
        Minutes.

    Returns
    -------
    `int`
        Timestamp in seconds.
    """
    sec = 0

    if days is not None:
        if not isinstance(days, int):
            raise TypeError(f"'days' must be an instance of 'int' but not of {type(days)}")
        if days <= 0:
            raise ValueError("'days' must be positive")

        sec += 24*60*60 * days
    if hours is not None:
        if not isinstance(hours, int):
            raise TypeError(f"'hours' must be an instance of 'int' but not of {type(hours)}")
        if hours <= 0:
            raise ValueError("'hours' must be positive")

        sec += 60*60 * hours
    if minutes is not None:
        if not isinstance(minutes, int):
            raise TypeError(f"'minutes' must be an instance of 'int' but not of {type(minutes)}")
        if minutes <= 0:
            raise ValueError("'minutes' must be positive")

        sec += 60 * minutes

    return sec



"""
Module provides functions for loading different water distribution networks.
"""
import os

from ..simulation import ScenarioConfig, ScenarioSimulator, SensorConfig
from ..utils import get_temp_folder, download_if_necessary


def create_empty_sensor_config(f_inp: str) -> SensorConfig:
    """
    Creates an empty sensor configuration for a given .inp file.

    Parameters
    ----------
    f_inp : `str`
        Path to the .inp file.

    Returns
    -------
    :class:`~epyt_flow.simulation.sensor_config.SensorConfig`
        Sensor configuration.
    """
    with ScenarioSimulator(f_inp_in=f_inp) as sim:
        return sim.sensor_config


def get_default_hydraulic_options(flow_units_id: int = None) -> dict:
    """
    Gets standard hydraulic default options -- i.e. switch to pressure-driven analysis.

    Parameters
    ----------
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    `dict`
        Dictionary with default hydraulics options that can be passed to
        :func:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator.set_general_parameters`.
    """
    params = {}
    if flow_units_id is not None:
        params |= {"flow_units_id": flow_units_id}

    return params


def load_inp(f_in: str, include_empty_sensor_config: bool = True,
             flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads an .inp file and wraps it into a scenario configuration.

    Parameters
    ----------
    f_in : `str`
        Path to the .inp file.
    include_empty_sensor_config : `bool`, optional
        If True, an empty sensor configuration will be included in the returned
        scenario configuration.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Scenario configuration for the .inp file.
    """
    if not os.path.isfile(f_in):
        raise ValueError(f"Can not find {f_in}")

    if include_empty_sensor_config is True:
        return ScenarioConfig(f_inp_in=f_in, sensor_config=create_empty_sensor_config(f_inp=f_in),
                              general_params=get_default_hydraulic_options(flow_units_id))
    else:
        return ScenarioConfig(f_inp_in=f_in,
                              general_params=get_default_hydraulic_options(flow_units_id))


def load_net1(download_dir: str = get_temp_folder(), verbose: bool = True,
              flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the Net1 network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Net1 network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "Net1.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/Net1.inp"
    backup_urls = ["https://raw.githubusercontent.com/OpenWaterAnalytics/EPyT/refs/heads/dev/epyt/networks/asce-tf-wdst/Net1.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_net2(download_dir: str = get_temp_folder(), verbose: bool = True,
              flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the Net2 network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Net2 network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "Net2.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/Net2.inp"
    backup_urls = ["https://raw.githubusercontent.com/OpenWaterAnalytics/EPyT/refs/heads/dev/epyt/networks/asce-tf-wdst/Net2.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_net3(download_dir: str = get_temp_folder(), verbose: bool = True,
              flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the Net3 network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Net3 network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "Net3.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/Net3.inp"
    backup_urls = ["https://raw.githubusercontent.com/OpenWaterAnalytics/EPyT/refs/heads/dev/epyt/networks/asce-tf-wdst/Net3.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_net6(download_dir: str = get_temp_folder(), verbose: bool = True,
              flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the Net6 network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Net6 network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "Net6.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/Net6.inp"

    download_if_necessary(f_in, url, verbose)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_richmond(download_dir: str = get_temp_folder(), verbose: bool = True,
                  flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the Richmond network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Richmond network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "Richmond_standard.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/Richmond_standard.inp"
    backup_urls = ["https://raw.githubusercontent.com/KIOS-Research/EPANET-Benchmarks/refs/heads/master/collect-epanet-inp/Richmond_standard.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_micropolis(download_dir: str = get_temp_folder(), verbose: bool = True,
                    flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the MICROPOLIS network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        MICROPOLIS network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "MICROPOLIS_v1.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/MICROPOLIS_v1.inp"
    backup_urls = ["https://raw.githubusercontent.com/KIOS-Research/EPANET-Benchmarks/refs/heads/master/collect-epanet-inp/MICROPOLIS_v1.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_balerma(download_dir: str = get_temp_folder(), verbose: bool = True,
                 flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the Balerma network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Balerma network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "Balerma.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/Balerma.inp"
    backup_urls = ["https://raw.githubusercontent.com/KIOS-Research/EPANET-Benchmarks/refs/heads/master/collect-epanet-inp/Balerma.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_rural(download_dir: str = get_temp_folder(), verbose: bool = True,
               flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the Rural network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Rural network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "RuralNetwork.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/RuralNetwork.inp"
    backup_urls = ["https://raw.githubusercontent.com/KIOS-Research/EPANET-Benchmarks/refs/heads/master/collect-epanet-inp/RuralNetwork.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_bwsn1(download_dir: str = get_temp_folder(), verbose: bool = True,
               flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the BWSN-1 network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        BWSN-1 network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "BWSN_Network_1.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/BWSN_Network_1.inp"
    backup_urls = ["https://raw.githubusercontent.com/KIOS-Research/EPANET-Benchmarks/refs/heads/master/collect-epanet-inp/BWSN_Network_1.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_bwsn2(download_dir: str = get_temp_folder(), verbose: bool = True,
               flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the BWSN-2 network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        BWSN-2 network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "BWSN_Network_2.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/BWSN_Network_2.inp"
    backup_urls = ["https://raw.githubusercontent.com/KIOS-Research/EPANET-Benchmarks/refs/heads/master/collect-epanet-inp/BWSN_Network_2.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_anytown(download_dir: str = get_temp_folder(), verbose: bool = True,
                 flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the Anytown network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Anytown network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "Anytown.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/Anytown.inp"
    backup_urls = ["https://raw.githubusercontent.com/OpenWaterAnalytics/EPyT/refs/heads/dev/epyt/networks/asce-tf-wdst/Anytown.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_dtown(download_dir: str = get_temp_folder(), verbose: bool = True,
               flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the D-Town network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        D-Town network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "d-town.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/d-town.inp"
    backup_urls = ["https://raw.githubusercontent.com/KIOS-Research/EPANET-Benchmarks/refs/heads/master/collect-epanet-inp/d-town.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_ctown(download_dir: str = get_temp_folder(), verbose: bool = True,
               flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the C-Town network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        C-Town network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "CTOWN.INP")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/CTOWN.INP"

    download_if_necessary(f_in, url, verbose)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_kentucky(wdn_id: int = 1, download_dir: str = get_temp_folder(),
                  verbose: bool = True, flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the specified Kentucky network.

    Parameters
    ----------
    wdn_id : `int`, optional
        The ID (1-15) of the particular network.

        The default is wdn_id=1
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Kentucky network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    if not isinstance(wdn_id, int):
        raise ValueError("'wdn_id' must be an integer in [1, 15]")
    if wdn_id < 1 or wdn_id > 15:
        raise ValueError(f"Unknown network 'ky{wdn_id}.inp'")

    f_in = os.path.join(download_dir, f"ky{wdn_id}.inp")
    url = f"https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/ky{wdn_id}.inp"
    backup_urls = [f"https://raw.githubusercontent.com/OpenWaterAnalytics/EPyT/refs/heads/dev/epyt/networks/asce-tf-wdst/ky{wdn_id}.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    return load_inp(f_in, flow_units_id=flow_units_id)


def load_hanoi(download_dir: str = get_temp_folder(),
               include_default_sensor_placement: bool = False,
               verbose: bool = True, flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the Hanoi network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    include_default_sensor_placement : `bool`, optional
        If True, a default sensor placement will be included in the returned scenario configuration.

        The default is False
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Hanoi network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_in = os.path.join(download_dir, "Hanoi.inp")
    url = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/Hanoi.inp"
    backup_urls = ["https://raw.githubusercontent.com/OpenWaterAnalytics/EPyT/refs/heads/dev/epyt/networks/asce-tf-wdst/Hanoi.inp"]

    download_if_necessary(f_in, url, verbose, backup_urls)
    config = load_inp(f_in, flow_units_id=flow_units_id)

    if include_default_sensor_placement is True:
        sensor_config = config.sensor_config
        sensor_config.pressure_sensors = ["13", "16", "22", "30"]
        sensor_config.flow_sensors = ["1"]

        config = ScenarioConfig(scenario_config=config, sensor_config=sensor_config)

    return config


def load_ltown(download_dir: str = get_temp_folder(), use_realistic_demands: bool = False,
               include_default_sensor_placement: bool = False,
               verbose: bool = True, flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the L-TOWN_v2 network.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    use_realistic_demands : `bool`, optional
        If True, realistic demands from the BattLeDIM challenge will be included,
        toy demands will be included otherwise.

        The default is False
    include_default_sensor_placement : `bool`, optional
        If True, the L-TOWN default sensor placement will be included
        in the returned scenario configuration.

        The default is False
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        L-TOWN_v2 network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_inp = "L-TOWN_v2_Model.inp" if use_realistic_demands is False else "L-TOWN_v2_Real.inp"

    f_in = os.path.join(download_dir, f_inp)
    if not use_realistic_demands:
        url = "https://zenodo.org/records/4017659/files/L-TOWN.inp?download=1"
    else:
        url = "https://zenodo.org/records/4017659/files/L-TOWN_Real.inp?download=1"

    download_if_necessary(f_in, url, verbose)
    config = load_inp(f_in, flow_units_id=flow_units_id)

    if include_default_sensor_placement is True:
        sensor_config = config.sensor_config
        sensor_config.pressure_sensors = ["n54", "n105", "n114", "n163", "n188", "n229", "n288",
                                          "n296", "n332", "n342", "n410", "n415", "n429", "n458",
                                          "n469", "n495", "n506", "n516", "n519", "n549", "n613",
                                          "n636", "n644", "n679", "n722", "n726", "n740", "n752",
                                          "n769"]
        sensor_config.flow_sensors = ["p227", "p235"]
        sensor_config.tank_volume_sensors = ["T1"]
        sensor_config.demand_sensors = ["n1", "n2",	"n3", "n4", "n6", "n7",	"n8", "n9",	"n10",
                                        "n11", "n13", "n16", "n17", "n18", "n19", "n20", "n21",
                                        "n22",	"n23", "n24", "n25", "n26", "n27", "n28", "n29",
                                        "n30", "n31", "n32", "n33",	"n34", "n35", "n36", "n39",
                                        "n40", "n41", "n42", "n43",	"n44", "n45", "n343", "n344",
                                        "n345",	"n346",	"n347",	"n349",	"n350",	"n351",	"n352",
                                        "n353",	"n354",	"n355",	"n356",	"n357",	"n358",	"n360",
                                        "n361",	"n362",	"n364",	"n365",	"n366",	"n367", "n368",
                                        "n369",	"n370",	"n371",	"n372",	"n373",	"n374", "n375",
                                        "n376",	"n377",	"n378", "n379",	"n381",	"n382",	"n383",
                                        "n384",	"n385",	"n386", "n387",	"n388",	"n389"]

        config = ScenarioConfig(scenario_config=config, sensor_config=sensor_config)

    return config


def load_ltown_a(download_dir: str = get_temp_folder(), use_realistic_demands: bool = False,
                 include_default_sensor_placement: bool = False,
                 verbose: bool = True, flow_units_id: int = None) -> ScenarioConfig:
    """
    Loads (and downloads if necessary) the L-TOWN-A network (area "A" of the L-TOWN network).

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the directory where the .inp file is stored.

        The default is the OS-specific temporary directory (e.g. "C:\\\\temp", "/tmp/", etc.)
    use_realistic_demands : `bool`, optional
        If True, realistic demands from the BattLeDIM challenge will be included,
        toy demands will be included otherwise.

        The default is False
    include_default_sensor_placement : `bool`, optional
        If True, the L-TOWN default sensor placement will be included
        in the returned scenario configuration.

        The default is False
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading the file.

        The default is True.
    flow_units_id : `int`, optional
        Specifies the flow units to be used in this scenario.
        If None, the units from the .inp file will be used.

        Must be one of the following EPANET toolkit constants:

            - EN_CFS  = 0  (cubic foot/sec)
            - EN_GPM  = 1  (gal/min)
            - EN_MGD  = 2  (Million gal/day)
            - EN_IMGD = 3  (Imperial MGD)
            - EN_AFD  = 4  (ac-foot/day)
            - EN_LPS  = 5  (liter/sec)
            - EN_LPM  = 6  (liter/min)
            - EN_MLD  = 7  (Megaliter/day)
            - EN_CMH  = 8  (cubic meter/hr)
            - EN_CMD  = 9  (cubic meter/day)

        The default is None.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        L-TOWN-A network loaded into a scenario configuration that can be passed on to
        :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
    """
    f_inp = "L-TOWN_v2-A_Model.inp" if use_realistic_demands is False else "L-TOWN_v2-A_Real.inp"

    f_in = os.path.join(download_dir, f_inp)
    url = f"https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/Networks/{f_inp}"

    download_if_necessary(f_in, url, verbose)
    config = load_inp(f_in, flow_units_id=flow_units_id)

    if include_default_sensor_placement is True:
        sensor_config = config.sensor_config
        sensor_config.pressure_sensors = ["n54", "n105", "n114", "n163", "n188", "n229", "n288",
                                          "n296", "n332", "n342", "n410", "n415", "n429", "n458",
                                          "n469", "n495", "n506", "n516", "n519", "n549", "n613",
                                          "n636", "n644", "n679", "n722", "n726", "n740", "n752",
                                          "n769"]
        sensor_config.flow_sensors = ["p227", "p235"]

        config = ScenarioConfig(scenario_config=config, sensor_config=sensor_config)

    return config



"""
This module provides REST API handlers for model and sensor uncertainties of scenarios.
"""
import warnings
import falcon

from .handlers import ScenarioBaseHandler
from ...simulation import SensorNoise, ModelUncertainty


class ScenarioModelUncertaintyHandler(ScenarioBaseHandler):
    """
    Class for handling GET and POST requests concerning model uncertainty.
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets the model uncertainties of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_model_uncertainties = self.scenario_mgr.get(scenario_id).model_uncertainty
            self.send_json_response(resp, my_model_uncertainties)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Sets the model uncertainties of a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            model_uncertainty = self.load_json_data_from_request(req)
            if not isinstance(model_uncertainty, ModelUncertainty):
                self.send_json_parsing_error(resp)
                return

            self.scenario_mgr.get(scenario_id).model_uncertainty = model_uncertainty
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioSensorUncertaintyHandler(ScenarioBaseHandler):
    """
    Class for handling GET and POST requests concerning sensor uncertainty (i.e. noise).
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets the sensor uncertainty (i.e. noise) of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_sensor_noise = self.scenario_mgr.get(scenario_id).sensor_noise
            self.send_json_response(resp, my_sensor_noise)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Sets the sensor uncertainty (i.e. noise) of a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            sensor_noise = self.load_json_data_from_request(req)
            if not isinstance(sensor_noise, SensorNoise):
                self.send_json_parsing_error(resp)
                return

            self.scenario_mgr.get(scenario_id).sensor_noise = sensor_noise
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR



"""
Module contains a class for representing complex control rules as implemented in EPANET.
"""
from copy import deepcopy
from typing import Any
import numpy as np
from epyt.epanet import ToolkitConstants

from ...serialization import JsonSerializable, COMPLEX_CONTROL_ID, COMPLEX_CONTROL_CONDITION_ID, \
    COMPLEX_CONTROL_ACTION_ID, serializable


EN_R_AND = 2
EN_R_OR  = 3

EN_R_DEMAND    = 1
EN_R_HEAD      = 2
EN_R_LEVEL     = 3
EN_R_PRESSURE  = 4
EN_R_FLOW      = 5
EN_R_STATUS    = 6
EN_R_SETTING   = 7
EN_R_POWER     = 8
EN_R_TIME      = 9
EN_R_CLOCKTIME = 10
EN_R_FILLTIME  = 11
EN_R_DRAINTIME = 12

EN_R_EQ      = 0
EN_R_NEQ     = 1
EN_R_LEQ     = 2
EN_R_GEQ     = 3
EN_R_LESS    = 4
EN_R_GREATER = 5
EN_R_IS      = 6
EN_R_NOT     = 7
EN_R_BELOW   = 8
EN_R_ABOVE   = 9

EN_R_ACTION_SETTING       = -1
EN_R_ACTION_STATUS_OPEN   = 1
EN_R_ACTION_STATUS_CLOSED = 2
EN_R_ACTION_STATUS_ACTIVE = 3


@serializable(COMPLEX_CONTROL_CONDITION_ID, ".epytflow_complex_control_condition")
class RuleCondition(JsonSerializable):
    """
    Class representing a rule condition.

    Parameters
    ----------
    object_type_id : `int`
        ID of the object type.

        Must be one of the following EPANET constants:

            - EN_R_NODE   = 6
            - EN_R_LINK   = 7
            - EN_R_SYSTEM = 8
    object_id : `str`
        ID of the object (i.e. junction, pipe, link, tank, etc.).
    attribute_id : `int`
        Type ID of the object's attribute that is checked.

        Must be on of the following constants:

            - EN_R_DEMAND    = 1
            - EN_R_HEAD      = 2
            - EN_R_LEVEL     = 3
            - EN_R_PRESSURE  = 4
            - EN_R_FLOW      = 5
            - EN_R_STATUS    = 6
            - EN_R_SETTING   = 7
            - EN_R_TIME      = 9
            - EN_R_CLOCKTIME = 10
            - EN_R_FILLTIME  = 11
            - EN_R_DRAINTIME = 12
    relation_type_id : `int`
        ID of the type of comparison.

        Must be one of the following constants:

            - EN_R_EQ      = 0
            - EN_R_NEQ     = 1
            - EN_R_LEQ     = 2
            - EN_R_GEQ     = 3
            - EN_R_LESS    = 4
            - EN_R_GREATER = 5
            - EN_R_IS      = 6
            - EN_R_NOT     = 7
            - EN_R_BELOW   = 8
            - EN_R_ABOVE   = 9
    value : `Any`
        Value that is compared against.
    """
    def __init__(self, object_type_id: int, object_id: str, attribute_id: int,
                 relation_type_id: int, value: Any, **kwds):
        if not isinstance(object_type_id, int):
            raise TypeError("'object_type_id' must be an instance of 'int' " +
                            f"but not of '{type(object_type_id)}'")
        if object_type_id not in [ToolkitConstants.EN_R_NODE, ToolkitConstants.EN_R_LINK,
                                  ToolkitConstants.EN_R_SYSTEM]:
            raise ValueError(f"Invalid value '{object_type_id}' for 'object_type_id'")
        if not isinstance(object_id, str):
            raise TypeError("'object_id' must be an instance of 'str' " +
                            f"but not of '{type(object_id)}'")
        if not isinstance(attribute_id, int):
            raise TypeError("'attribute_id' must be an instance of 'int' " +
                            f"but not of '{type(attribute_id)}'")
        if attribute_id not in [EN_R_DEMAND, EN_R_HEAD, EN_R_LEVEL, EN_R_PRESSURE,
                                EN_R_FLOW, EN_R_STATUS, EN_R_SETTING, EN_R_POWER, EN_R_TIME,
                                EN_R_CLOCKTIME, EN_R_FILLTIME, EN_R_DRAINTIME]:
            raise ValueError(f"Invalid value '{attribute_id}' for 'attribute_id'")
        if not isinstance(relation_type_id, int):
            raise TypeError("'relation_type_id' must be an instance of 'int' " +
                            f"but not of '{type(relation_type_id)}'")
        if relation_type_id not in [EN_R_EQ, EN_R_NEQ, EN_R_LEQ, EN_R_GEQ, EN_R_LESS, EN_R_GREATER,
                                    EN_R_IS, EN_R_NOT, EN_R_BELOW, EN_R_ABOVE]:
            raise ValueError(f"Invalid value '{relation_type_id}' for 'relation_type_id'")

        self.__object_type_id = object_type_id
        self.__object_id = object_id
        self.__attribute_id = attribute_id
        self.__relation_type_id = relation_type_id
        self.__value = value

        super().__init__(**kwds)

    @property
    def object_type_id(self) -> int:
        """
        Returns the ID of the object type.

        Will be one of the following EPANET constants:

            - EN_R_NODE   = 6
            - EN_R_LINK   = 7
            - EN_R_SYSTEM = 8

        Returns
        -------
        `int`
            ID of the object type..
        """
        return self.__object_type_id

    @property
    def object_id(self) -> str:
        """
        Returns the ID of the object (i.e. junction, pipe, link, tank, etc.).

        Returns
        -------
        `str`
            ID of the object.
        """
        return self.__object_id

    @property
    def attribute_id(self) -> int:
        """
        Returns the type ID of the object's attribute that is checked.

        Will be one of the following constants:

            - EN_R_DEMAND    = 1
            - EN_R_HEAD      = 2
            - EN_R_LEVEL     = 3
            - EN_R_PRESSURE  = 4
            - EN_R_FLOW      = 5
            - EN_R_STATUS    = 6
            - EN_R_SETTING   = 7
            - EN_R_TIME      = 9
            - EN_R_CLOCKTIME = 10
            - EN_R_FILLTIME  = 11
            - EN_R_DRAINTIME = 12

        Returns
        -------
        `int`
            Type ID of the object's attribute that is checked.
        """
        return self.__attribute_id

    @property
    def relation_type_id(self) -> int:
        """
        Returns the ID of the type of comparison.

        Will be one of the following constants:

            - EN_R_EQ      = 0
            - EN_R_NEQ     = 1
            - EN_R_LEQ     = 2
            - EN_R_GEQ     = 3
            - EN_R_LESS    = 4
            - EN_R_GREATER = 5
            - EN_R_IS      = 6
            - EN_R_NOT     = 7
            - EN_R_BELOW   = 8
            - EN_R_ABOVE   = 9

        Returns
        -------
        `int`
            ID of the type of comparison.
        """
        return self.__relation_type_id

    @property
    def value(self) -> Any:
        """
        Returns the value that is compared against.

        Returns
        -------
        `Any`
            Value that is compared against.
        """
        return self.__value

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"object_type_id": self.__object_type_id,
                                           "object_id": self.__object_id,
                                           "attribute_id": self.__attribute_id,
                                           "relation_type_id": self.__relation_type_id,
                                           "value": self.__value}

    def __eq__(self, other) -> bool:
        return self.__object_type_id == other.object_type_id and \
            self.__object_id == other.object_id and self.__attribute_id == other.attribute_id and \
            self.__relation_type_id == other.relation_type_id and self.__value == other.value

    def __str__(self) -> str:
        desc = ""

        if self.__attribute_id == EN_R_DEMAND:
            if self.__object_type_id == ToolkitConstants.EN_R_NODE:
                desc += f"JUNCTION {self.__object_id} DEMAND "
            elif self.__object_type_id == ToolkitConstants.EN_R_SYSTEM:
                desc += "SYSTEM DEMAND "
        elif self.__attribute_id == EN_R_HEAD:
            desc += f"JUNCTION {self.__object_id} HEAD "
        elif self.__attribute_id == EN_R_LEVEL:
            desc += f"TANK {self.__object_id} LEVEL "
        elif self.__attribute_id == EN_R_PRESSURE:
            desc += f"JUNCTION {self.__object_id} PRESSURE "
        elif self.__attribute_id == EN_R_FLOW:
            desc += f"LINK {self.__object_id} FLOW "
        elif self.__attribute_id == EN_R_STATUS:
            desc += f"LINK {self.__object_id} STATUS "
        elif self.__attribute_id == EN_R_SETTING:
            desc += f"LINK {self.__object_id} SETTING "
        elif self.__attribute_id == EN_R_TIME:
            desc += "SYSTEM TIME "
        elif self.__attribute_id == EN_R_CLOCKTIME:
            desc += "SYSTEM CLOCKTIME "
        elif self.__attribute_id == EN_R_FILLTIME:
            desc += f"TANK {self.__object_id} FILLTIME "
        elif self.__attribute_id == EN_R_DRAINTIME:
            desc += f"TANK {self.__object_id} DRAINTIME "

        if self.__relation_type_id == EN_R_EQ:
            desc += "= "
        elif self.__relation_type_id == EN_R_IS:
            desc += "IS "
        elif self.__relation_type_id == EN_R_NOT:
            desc += "IS NOT "
        elif self.__relation_type_id == EN_R_LEQ:
            desc += "<= "
        elif self.__relation_type_id == EN_R_GEQ:
            desc += ">= "
        elif self.__relation_type_id == EN_R_ABOVE:
            desc += "ABOVE "
        elif self.__relation_type_id == EN_R_BELOW:
            desc += "BELOW "
        elif self.__relation_type_id == EN_R_LESS:
            desc += "< "
        elif self.__relation_type_id == EN_R_GREATER:
            desc += "> "

        desc += str(self.__value)

        return desc


@serializable(COMPLEX_CONTROL_ACTION_ID, ".epytflow_complex_control_action")
class RuleAction(JsonSerializable):
    """
    Class representing a rule action.

    Parameters
    ----------
    link_type_id : `int`
        Link type ID.

        Must be one of following EPANET constants:

            - EN_CVPIPE = 0
            - EN_PIPE = 1
            - EN_PUMP = 2
            - EN_PRV = 3
            - EN_PSV = 4
            - EN_PBV = 5
            - EN_FCV = 6
            - EN_TCV = 7
            - EN_GPV = 8
    link_id : `str`
        Link ID.
    action_type_id : `int`
        Type ID of the action.

        Must be one of the following constants:

            - EN_R_ACTION_SETTING       = -1
            - EN_R_ACTION_STATUS_OPEN   = 1
            - EN_R_ACTION_STATUS_CLOSED = 2
            - EN_R_ACTION_STATUS_ACTIVE = 3
    action_value : `Any`
        Value of the acton (e.g. pump speed).
        Only relevant if action_type_id = EN_R_SETTING, will be ignored in all other cases.
    """
    def __init__(self, link_type_id: int, link_id: str, action_type_id: int, action_value: Any,
                 **kwds):
        if not isinstance(link_type_id, int):
            raise TypeError("'link_type_id' must be an istanace of 'int' " +
                            f"but not of '{type(link_type_id)}'")
        if link_type_id not in [ToolkitConstants.EN_CVPIPE, ToolkitConstants.EN_PIPE,
                                ToolkitConstants.EN_PUMP, ToolkitConstants.EN_PRV,
                                ToolkitConstants.EN_PSV, ToolkitConstants.EN_PBV,
                                ToolkitConstants.EN_FCV, ToolkitConstants.EN_TCV,
                                ToolkitConstants.EN_GPV]:
            raise ValueError(f"Invalid value '{link_type_id}' for 'link_type_id'")
        if not isinstance(link_id, str):
            raise TypeError("'link_id' must be an instance of 'str' " +
                            f"but not of '{type(link_id)}'")
        if not isinstance(action_type_id, int):
            raise TypeError("'action_type_id' must be an instance of 'int' " +
                            f"but not of '{type(action_type_id)}'")
        if action_type_id not in [EN_R_ACTION_SETTING, EN_R_ACTION_STATUS_OPEN,
                                  EN_R_ACTION_STATUS_CLOSED, EN_R_ACTION_STATUS_ACTIVE]:
            raise ValueError(f"Invalid value '{action_type_id}' for 'action_type_id'")

        self.__link_type_id = link_type_id
        self.__link_id = link_id
        self.__action_type_id = action_type_id
        self.__action_value = action_value

        super().__init__(**kwds)

    @property
    def link_type_id(self) -> int:
        """
        Returns the link type ID.

        Will be one of the following EPANET constants:

            - EN_CVPIPE = 0
            - EN_PIPE = 1
            - EN_PUMP = 2
            - EN_PRV = 3
            - EN_PSV = 4
            - EN_PBV = 5
            - EN_FCV = 6
            - EN_TCV = 7
            - EN_GPV = 8

        Returns
        -------
        `int`
            Link type ID.
        """
        return self.__link_type_id

    @property
    def link_id(self) -> str:
        """
        Returns the link ID.

        Returns
        -------
        `str`
            Link ID.
        """
        return self.__link_id

    @property
    def action_type_id(self) -> int:
        """
        Returns the type ID of the action.

        Will be one of the following constants:

            - EN_R_ACTION_SETTING       = -1
            - EN_R_ACTION_STATUS_OPEN   = 1
            - EN_R_ACTION_STATUS_CLOSED = 2
            - EN_R_ACTION_STATUS_ACTIVE = 3

        Returns
        -------
        `int`
            Type ID of the action.
        """
        return self.__action_type_id

    @property
    def action_value(self) -> Any:
        """
        Returns the value of the acton (e.g. pump speed).
        Only relevant if action_type_id = EN_R_SETTING.

        Returns
        -------
        `Any`
            Value of the action.
        """
        return self.__action_value

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"link_type_id": self.__link_type_id,
                                           "link_id": self.__link_id,
                                           "action_type_id": self.__action_type_id,
                                           "action_value": self.__action_value}

    def __eq__(self, other) -> bool:
        return self.__link_type_id == other.link_type_id and \
            self.__link_id == other.link_id and \
            self.__action_type_id == other.action_type_id and \
            self.__action_value == other.action_value

    def __str__(self) -> str:
        desc = ""

        if self.__link_type_id in [ToolkitConstants.EN_CVPIPE, ToolkitConstants.EN_PIPE]:
            desc += "PIPE "
        elif self.__link_type_id == ToolkitConstants.EN_PUMP:
            desc += "PUMP "
        else:
            desc += "VALVE "

        desc += f"{self.__link_id} "

        if self.__action_type_id == EN_R_ACTION_SETTING:
            desc += f"SETTING IS {self.__action_value}"
        elif self.__action_type_id == EN_R_ACTION_STATUS_OPEN:
            desc += "STATUS IS OPEN"
        elif self.__action_type_id == EN_R_ACTION_STATUS_CLOSED:
            desc += "STATUS IS CLOSED"
        elif self.action_type_id == EN_R_ACTION_STATUS_ACTIVE:
            desc += "STATUS IS ACTIVE"

        return desc


@serializable(COMPLEX_CONTROL_ID, ".epytflow_complex_control")
class ComplexControlModule(JsonSerializable):
    """
    Class representing a complex control module (i.e. IF-THEN-ELSE rule) as implemented in EPANET.

    Parameters
    ----------
    rule_id : `str`
        ID of the rule.
    condition_1 : :class:`~epyt_flow.simulation.scada.complex_control.RuleCondition`
        First condition of this rule.
    additional_conditions : list[tuple[int, :class:`~epyt_flow.simulation.scada.complex_control.RuleCondition`]]
        List of (optional) additional conditions incl. their conjunction operator
        (must be either EN_R_AND = 2 or EN_R_OR  = 3).

        Empty list if there are no additional conditions.
    actions : list[:class:`~epyt_flow.simulation.scada.complex_control.RuleAction`]
        List of actions that are applied if the conditions are met.
        Must contain at least one action.
    else_actions : list[:class:`~epyt_flow.simulation.scada.complex_control.RuleAction`]
        List of actions that are applied if the conditions are NOT met.
    priority : `int`
        Priority of this control rule.
    """
    def __init__(self, rule_id: str, condition_1: RuleCondition,
                 additional_conditions: list[tuple[int, RuleCondition]], actions: list[RuleAction],
                 else_actions: list[RuleAction], priority: int, **kwds):
        if not isinstance(rule_id, str):
            raise TypeError(f"'rule_id' must be an instance of 'str' but not of '{type(rule_id)}'")
        if not isinstance(condition_1, RuleCondition):
            raise TypeError("'condition_1' must be an instance of " +
                            "'epyt_flow.simulation.scada.RuleCondition' " +
                            f"but not of '{type(condition_1)}'")
        if not isinstance(additional_conditions, list) or \
                any(not isinstance(condition, tuple) for condition in additional_conditions) or \
                any(not isinstance(condition[0], int) or condition[0] not in [EN_R_AND, EN_R_OR] or
                    not isinstance(condition[1], RuleCondition)
                    for condition in additional_conditions):
            raise TypeError("'additional_conditions' must be a list of " +
                            "'tuple[int, epyt_flow.simulation.scada.RuleCondition]' instances")
        if not isinstance(actions, list) or any(not isinstance(action, RuleAction)
                                                for action in actions):
            raise TypeError("'actions' must be a list of " +
                            "'epyt_flow.simulation.scada.RuleAction' instances")
        if len(actions) == 0:
            raise ValueError("'actions' must contain at least one action")
        if not isinstance(else_actions, list) or any(not isinstance(action, RuleAction)
                                                     for action in actions):
            raise TypeError("'else_actions' must be a list of " +
                            "'epyt_flow.simulation.scada.RuleAction' instances")
        if not isinstance(priority, int) or priority < 0:
            raise TypeError("'priority' must be a non-negative integer")

        self.__rule_id = rule_id
        self.__condition_1 = condition_1
        self.__additional_conditions = additional_conditions
        self.__actions = actions
        self.__else_actions = else_actions
        self.__priority = priority

        super().__init__(**kwds)

    @property
    def rule_id(self) -> str:
        """
        Returns the ID of this control rule.

        Returns
        -------
        `str`
            ID of this control rule.
        """
        return self.__rule_id

    @property
    def condition_1(self) -> RuleCondition:
        """
        Returns the first condition of this rule.

        Returns
        -------
        :class:`~epyt_flow.simulation.scada.complex_control.RuleCondition`
            First condition of this rule.
        """
        return deepcopy(self.__condition_1)

    @property
    def additional_conditions(self) -> list[tuple[int, RuleCondition]]:
        """
        Returns the list of (optional) additional conditions incl. their conjunction operator.
        Empty list if there are no additional conditions.

        Returns
        -------
        list[tuple[int, :class:`~epyt_flow.simulation.scada.complex_control.RuleCondition`]]
            List of (optional) additional conditions incl. their conjunction operator.
        """
        return deepcopy(self.__additional_conditions)

    @property
    def actions(self) -> list[RuleAction]:
        """
        Returns the list of actions that are applied if the conditions are met.

        Returns
        -------
        list[:class:`~epyt_flow.simulation.scada.complex_control.RuleAction`]
            List of actions that are applied if the conditions are met.
        """
        return deepcopy(self.__actions)

    @property
    def else_actions(self) -> list[RuleAction]:
        """
        Returns the list of actions that are applied if the conditions are NOT met.

        Returns
        -------
        list[:class:`~epyt_flow.simulation.scada.complex_control.RuleAction`]
            List of actions that are applied if the conditions are NOT met.
        """
        return deepcopy(self.__else_actions)

    @property
    def priority(self) -> int:
        """
        Returns the priority of this control rule.

        Returns
        -------
        `int`
            Priority of this control rule.
        """
        return self.__priority

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"rule_id": self.__rule_id,
                                           "condition_1": self.__condition_1,
                                           "additional_conditions": self.__additional_conditions,
                                           "actions": self.__actions,
                                           "else_actions": self.__else_actions,
                                           "priority": self.__priority}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self.__rule_id == other.rule_id and \
            self.__priority == other.priority and self.__condition_1 == other.condition_1 and \
            np.all(self.__additional_conditions == other.additional_conditions) and \
            np.all(self.__actions == other.actions) and \
            np.all(self.__else_actions == other.else_actions)

    def __str__(self) -> str:
        desc = ""

        desc += f"RULE {self.__rule_id}\n"
        desc += f"IF {self.__condition_1} "
        for op, action in self.__additional_conditions:
            if op == EN_R_AND:
                desc += "\nAND "
            elif op == EN_R_OR:
                desc += "\nOR "

            desc += f"{action} "
        desc += "\nTHEN "
        desc += "\nAND ".join(str(action) for action in self.__actions)
        if len(self.__else_actions) != 0:
            desc += "\nELSE " + "\nAND ".join(str(action) for action in self.__else_actions)

        desc += f"\nPRIORITY {self.__priority}"

        return desc



"""
Module provides some helper functions regarding the implementation of uncertainty.
"""
import numpy as np
from numpy.random import Generator
from scipy.ndimage import gaussian_filter1d


def smoothing(pattern: np.ndarray, sigma: float = 10.) -> np.ndarray:
    """
    Smoothes a given pattern by applying a Gaussian filter.

    Parameters
    ----------
    pattern : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        The original pattern
    sigma : `float`, optional
        Standard deviation for the Gaussian filter.

        The default is 10.

    Returns
    -------
    `numpy.ndarray`
        The smoothed pattern.
    """
    return gaussian_filter1d(pattern, sigma=sigma)


def scale_to_range(pattern: np.ndarray, min_value: float, max_value: float) -> np.ndarray:
    """
    Scales a given pattern to an interval.

    Parameters
    ----------
    pattern : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        The pattern to be scaled.
    min_value : `float`
        Lower bound of the pattern.
    max_value : `float`
        Upper bound of the pattern.

    Returns
    -------
    `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        The scaled pattern.
    """
    if min_value is None or max_value is None:
        return pattern

    min_pattern_val = np.min(pattern)
    max_pattern_val = np.max(pattern)

    return [(x - min_pattern_val) / (max_pattern_val - min_pattern_val) * (max_value - min_value)
            + min_value for x in pattern]


def generate_random_gaussian_noise(n_samples: int, np_rand_gen: Generator = np.random.default_rng()
                                   ) -> np.ndarray:
    """
    Generates Gaussian noise using a random mean ([0,1]) and random standard deviation ([0,1]).

    Parameters
    ----------
    n_samples : `int`
        Number of random samples.
    np_rand_generator : `numpy.random.Generator <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.Generator>`_, optional
        The random number generator that is going to be used.

        The default is the default BitGenerator (PCG64) as constructed by
        `numpy.random.default_rng() <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.default_rng>`_.

    Returns
    -------
    `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Gaussian noise.
    """
    return np_rand_gen.normal(np.random.rand(), np.random.rand(), size=n_samples)


def generate_deep_random_gaussian_noise(n_samples: int, mean: float = None,
                                        np_rand_gen: Generator = np.random.default_rng()
                                        ) -> np.ndarray:
    """
    Generates random Gaussian noise where the standard deviations (and mean) are changing over time.

    Parameters
    ----------
    n_samples : `int`
        Number of random samples.
    mean : `float`, optional
        Fixed mean at all points in time.
        If None, random means are generated.

        The default is None.
    np_rand_generator : `numpy.random.Generator <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.Generator>`_, optional
        The random number generator that is going to be used.

        The default is the default BitGenerator (PCG64) as constructed by
        `numpy.random.default_rng() <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.default_rng>`_.

    Returns
    -------
    `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Random Gaussian noise.
    """
    noise = []

    if mean is None:
        mean = create_deep_random_pattern(n_samples, min_value=-1., max_value=1.,
                                          np_rand_gen=np_rand_gen)
    else:
        mean = [mean] * n_samples
    rand_std = create_deep_random_pattern(n_samples, np_rand_gen=np_rand_gen)
    noise = np.array([np_rand_gen.normal(m, s) for m, s in zip(mean, rand_std)])

    return noise


def create_deep_random_pattern(n_samples: int, min_value: float = 0., max_value: float = 1.,
                               init_value: float = None,
                               np_rand_gen: Generator = np.random.default_rng()) -> np.ndarray:
    """
    Generates a random pattern.

    Parameters
    ----------
    n_samples : `int`
        Number of random samples -- i.e. length of the pattern.
    min_value : `float`, optional
        Lower bound of the pattern.

        The default is zero.
    max_value : `float`, optional
        Upper bound of the pattern.

        The default is one.
    init_value : `float`, optional
        Value of the first sample in the pattern.
        If None, a random value is used.

        The default is None.
    np_rand_generator : `numpy.random.Generator <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.Generator>`_, optional
        The random number generator that is going to be used.

        The default is the default BitGenerator (PCG64) as constructed by
        `numpy.random.default_rng() <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.default_rng>`_.

    Returns
    -------
    `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Random pattern.
    """
    pattern = []
    start_value = init_value

    while len(pattern) < n_samples:
        if len(pattern) != 0:
            start_value = pattern[-1]

        pattern += _create_deep_random_pattern(start_value, min_value=min_value,
                                               max_value=max_value, np_rand_gen=np_rand_gen)

    pattern = pattern[:n_samples]

    # Scaling to value range
    pattern = scale_to_range(pattern, min_value, max_value)

    return np.array(pattern)


def _create_deep_random_pattern(start_value: float = None, min_length: int = 2, max_length: int = 5,
                                min_value: float = None, max_value: float = None,
                                np_rand_gen: Generator = np.random.default_rng()) -> np.ndarray:
    """
    Generates a random pattern of random length.

    Parameters
    ----------
    start_value : `float`, optional
        First value in the pattern.
        If None, a random value is used.

        The default is None.
    min_length : `int`, optional
        Minium length of the pattern.

        The default is 2.
    max_length : `int`
        Maximum length of the pattern.

        The default is 5.
    min_value : `float`, optional
        Lower bound of the pattern.

        The default is zero.
    max_value : `float`, optional
        Upper bound of the pattern.

        The default is one.
    np_rand_generator : `numpy.random.Generator <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.Generator>`_, optional
        The random number generator that is going to be used.

        The default is the default BitGenerator (PCG64) as constructed by
        `numpy.random.default_rng() <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.default_rng>`_.

    Returns
    -------
    `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Random pattern.
    """
    pattern = []

    # Random parameters of pattern
    if start_value is None:
        start_value = np_rand_gen.random()
    length = np_rand_gen.integers(low=min_length, high=max_length)
    vec = np_rand_gen.choice([-.1, .1])

    # Generate pattern
    cur_value = start_value
    pattern.append(start_value)

    for _ in range(length):
        cur_value = cur_value + np_rand_gen.random() * vec
        pattern.append(cur_value)
        if min_value is not None and max_value is not None:
            if cur_value < min_value:
                vec = .1
            elif cur_value > max_value:
                vec = -.1

    return pattern
